#include "row_acl.h"
#include <access/gist.h>
#include <access/rtree.h>
#include <ctype.h>

PG_FUNCTION_INFO_V1(row_acl_in);
PG_FUNCTION_INFO_V1(row_acl_out);
PG_FUNCTION_INFO_V1(row_acl_eq);
PG_FUNCTION_INFO_V1(row_acl_is_intersects);
PG_FUNCTION_INFO_V1(row_acl_contains);
PG_FUNCTION_INFO_V1(row_acl_within);
PG_FUNCTION_INFO_V1(int_to_row_acl);
PG_FUNCTION_INFO_V1(row_acl_to_int);
PG_FUNCTION_INFO_V1(row_acl_intersect);
PG_FUNCTION_INFO_V1(row_acl_union);
PG_FUNCTION_INFO_V1(row_acl_subtract);
PG_FUNCTION_INFO_V1(row_acl_set);
PG_FUNCTION_INFO_V1(row_acl_set_alt);
PG_FUNCTION_INFO_V1(gist_row_acl_compress);
PG_FUNCTION_INFO_V1(gist_row_acl_decompress);
PG_FUNCTION_INFO_V1(gist_row_acl_same);
PG_FUNCTION_INFO_V1(gist_row_acl_penalty);
PG_FUNCTION_INFO_V1(gist_row_acl_union);
PG_FUNCTION_INFO_V1(gist_row_acl_consistent);
PG_FUNCTION_INFO_V1(gist_row_acl_picksplit);

Datum row_acl_in(PG_FUNCTION_ARGS);
Datum row_acl_out(PG_FUNCTION_ARGS);
Datum row_acl_eq(PG_FUNCTION_ARGS);
Datum row_acl_is_intersects(PG_FUNCTION_ARGS);
Datum row_acl_contains(PG_FUNCTION_ARGS);
Datum row_acl_within(PG_FUNCTION_ARGS);
Datum int_to_row_acl(PG_FUNCTION_ARGS);
Datum row_acl_to_int(PG_FUNCTION_ARGS);
Datum row_acl_intersect(PG_FUNCTION_ARGS);
Datum row_acl_union(PG_FUNCTION_ARGS);
Datum row_acl_subtract(PG_FUNCTION_ARGS);
Datum row_acl_set(PG_FUNCTION_ARGS);
Datum row_acl_set_alt(PG_FUNCTION_ARGS);
Datum gist_row_acl_compress(PG_FUNCTION_ARGS);
Datum gist_row_acl_decompress(PG_FUNCTION_ARGS);
Datum gist_row_acl_same(PG_FUNCTION_ARGS);
Datum gist_row_acl_penalty(PG_FUNCTION_ARGS);
Datum gist_row_acl_union(PG_FUNCTION_ARGS);
Datum gist_row_acl_consistent(PG_FUNCTION_ARGS);
Datum gist_row_acl_picksplit(PG_FUNCTION_ARGS);

row_acl inner_row_acl_set(row_acl a, text *t);

Datum
row_acl_in(PG_FUNCTION_ARGS) {
  char *str = (char*)PG_GETARG_CSTRING(0);
  row_acl* result;
  int acl_error = 0;
  size_t sz = 0;
  text* t = NULL;
  if(str == NULL) PG_RETURN_NULL();
  result = (row_acl*)palloc(sizeof(row_acl));
  *result = ROW_ACL_NOTHING;
  if(strcmp(str, "") == 0) PG_RETURN_POINTER(result);
  if (str[0]=='u' || str[0]=='g' || str[0]=='o' || str[0]=='a') {
  /* Permissions in the "ug+rwdi,g-dw,o+r" form */
    sz = strlen(str);
    t = (text*)palloc(VARHDRSZ + sz);
    memcpy(VARDATA(t), str, sz);
    VARATT_SIZEP(t) = VARHDRSZ + sz;
    *result = inner_row_acl_set(ROW_ACL_NOTHING, t);
  } else {
  /* User permissions */
    switch (str[0])  { case 'r': *result |= ROW_ACL_USER_READ; break; case '-': break; default: acl_error = 1; }
    switch (str[1])  { case 'w': *result |= ROW_ACL_USER_WRITE; break; case '-': break; default: acl_error = 1; }
    switch (str[2])  { case 'd': *result |= ROW_ACL_USER_DELETE; break; case '-': break; default: acl_error = 1; }
    switch (str[3])  { case 'i': *result |= ROW_ACL_USER_INSERT; break; case '-': break; default: acl_error = 1; }
    switch (str[4])  { case 'x': *result |= ROW_ACL_USER_EXEC; break; case '-': break; default: acl_error = 1; }
  /* Group permissions */
    switch (str[5])  { case 'r': *result |= ROW_ACL_GROUP_READ; break; case '-': break; default: acl_error = 1; }
    switch (str[6])  { case 'w': *result |= ROW_ACL_GROUP_WRITE; break; case '-': break; default: acl_error = 1; }
    switch (str[7])  { case 'd': *result |= ROW_ACL_GROUP_DELETE; break; case '-': break; default: acl_error = 1; }
    switch (str[8])  { case 'i': *result |= ROW_ACL_GROUP_INSERT; break; case '-': break; default: acl_error = 1; }
    switch (str[9])  { case 'x': *result |= ROW_ACL_GROUP_EXEC; break; case '-': break; default: acl_error = 1; }
  /* Other permissions */
    switch (str[10]) { case 'r': *result |= ROW_ACL_OTHER_READ; break; case '-': break; default: acl_error = 1; }
    switch (str[11]) { case 'w': *result |= ROW_ACL_OTHER_WRITE; break; case '-': break; default: acl_error = 1; }
    switch (str[12]) { case 'd': *result |= ROW_ACL_OTHER_DELETE; break; case '-': break; default: acl_error = 1; }
    switch (str[13]) { case 'i': *result |= ROW_ACL_OTHER_INSERT; break; case '-': break; default: acl_error = 1; }
    switch (str[14]) { case 'x': *result |= ROW_ACL_OTHER_EXEC; break; case '-': break; default: acl_error = 1; }
  }
  if (acl_error) { elog(ERROR,"row_acl_in: incorrect access mode '%s'", str); PG_RETURN_NULL(); }
  PG_RETURN_POINTER(result);
}

Datum
row_acl_out(PG_FUNCTION_ARGS) {
  row_acl *ma = (row_acl*)PG_GETARG_POINTER(0);
  char* result;
  if(ma == NULL) PG_RETURN_NULL();
  result = (char*)palloc(ROW_ACL_STR_LENGTH);
  result[0] = 0;
/* User permissions */
  if (*ma & ROW_ACL_USER_READ) strcat(result, "r"); else strcat(result, "-");
  if (*ma & ROW_ACL_USER_WRITE) strcat(result, "w"); else strcat(result, "-");
  if (*ma & ROW_ACL_USER_DELETE) strcat(result, "d"); else strcat(result, "-");
  if (*ma & ROW_ACL_USER_INSERT) strcat(result, "i"); else strcat(result, "-");
  if (*ma & ROW_ACL_USER_EXEC) strcat(result, "x"); else strcat(result, "-");
/* Group permissions */
  if (*ma & ROW_ACL_GROUP_READ) strcat(result, "r"); else strcat(result, "-");
  if (*ma & ROW_ACL_GROUP_WRITE) strcat(result, "w"); else strcat(result, "-");
  if (*ma & ROW_ACL_GROUP_DELETE) strcat(result, "d"); else strcat(result, "-");
  if (*ma & ROW_ACL_GROUP_INSERT) strcat(result, "i"); else strcat(result, "-");
  if (*ma & ROW_ACL_GROUP_EXEC) strcat(result, "x"); else strcat(result, "-");
/* Other permissions */
  if (*ma & ROW_ACL_OTHER_READ) strcat(result, "r"); else strcat(result, "-");
  if (*ma & ROW_ACL_OTHER_WRITE) strcat(result, "w"); else strcat(result, "-");
  if (*ma & ROW_ACL_OTHER_DELETE) strcat(result, "d"); else strcat(result, "-");
  if (*ma & ROW_ACL_OTHER_INSERT) strcat(result, "i"); else strcat(result, "-");
  if (*ma & ROW_ACL_OTHER_EXEC) strcat(result, "x"); else strcat(result, "-");
  PG_RETURN_CSTRING(result);
}

Datum
row_acl_eq(PG_FUNCTION_ARGS) {
  row_acl* a = (row_acl*)PG_GETARG_POINTER(0);
  row_acl* b = (row_acl*)PG_GETARG_POINTER(1);
  if (*a == *b) PG_RETURN_BOOL(true);
  else PG_RETURN_BOOL(false);
}

Datum
row_acl_is_intersects(PG_FUNCTION_ARGS) {
  row_acl* a = (row_acl*)PG_GETARG_POINTER(0);
  row_acl* b = (row_acl*)PG_GETARG_POINTER(1);
  if (*a & *b) PG_RETURN_BOOL(true);
  else PG_RETURN_BOOL(false);
}

Datum
row_acl_contains(PG_FUNCTION_ARGS) {
  row_acl* a = (row_acl*)PG_GETARG_POINTER(0);
  row_acl* b = (row_acl*)PG_GETARG_POINTER(1);
  if ((*a & *b) == *b) PG_RETURN_BOOL(true);
  else PG_RETURN_BOOL(false);
}

Datum
row_acl_within(PG_FUNCTION_ARGS) {
  row_acl* a = (row_acl*)PG_GETARG_POINTER(0);
  row_acl* b = (row_acl*)PG_GETARG_POINTER(1);
  if ((*a & *b) == *a) PG_RETURN_BOOL(true);
  else PG_RETURN_BOOL(false);
}

Datum
int_to_row_acl(PG_FUNCTION_ARGS) {
  int32 mode = PG_GETARG_INT32(0);
  row_acl* result;
  result = (row_acl*)palloc(sizeof(row_acl));
  *result = mode;
  PG_RETURN_POINTER(result);
}

Datum
row_acl_to_int(PG_FUNCTION_ARGS) {
  row_acl* a = (row_acl*)PG_GETARG_POINTER(0);
  int32 result;
  result = *a;
  PG_RETURN_INT32(result);
}

Datum
row_acl_intersect(PG_FUNCTION_ARGS) {
  row_acl* a = (row_acl*)PG_GETARG_POINTER(0);
  row_acl* b = (row_acl*)PG_GETARG_POINTER(1);
  row_acl* result;
  result = (row_acl*)palloc(sizeof(row_acl));
  *result = *a & *b;
  PG_RETURN_POINTER(result);
}

Datum
row_acl_union(PG_FUNCTION_ARGS) {
  row_acl* a = (row_acl*)PG_GETARG_POINTER(0);
  row_acl* b = (row_acl*)PG_GETARG_POINTER(1);
  row_acl* result;
  result = (row_acl*)palloc(sizeof(row_acl));
  *result = *a | *b;
  PG_RETURN_POINTER(result);
}

Datum
row_acl_subtract(PG_FUNCTION_ARGS) {
  row_acl* a = (row_acl*)PG_GETARG_POINTER(0);
  row_acl* b = (row_acl*)PG_GETARG_POINTER(1);
  row_acl* result;
  result = (row_acl*)palloc(sizeof(row_acl));
  *result = *a ^ *b;
  PG_RETURN_POINTER(result);
}

Datum
row_acl_set(PG_FUNCTION_ARGS) {
  row_acl *a = (row_acl*)PG_GETARG_POINTER(0);
  text *t = PG_GETARG_TEXT_P(1);
  row_acl *result = (row_acl*)palloc(sizeof(row_acl));

  *result = inner_row_acl_set(*a, t);
  PG_RETURN_POINTER(result);
}

Datum
row_acl_set_alt(PG_FUNCTION_ARGS) {
  text *t = PG_GETARG_TEXT_P(0);
  row_acl *result = (row_acl*)palloc(sizeof(row_acl));

  *result = inner_row_acl_set(ROW_ACL_NOTHING, t);
  PG_RETURN_POINTER(result);
}

row_acl
inner_row_acl_set(row_acl a, text *t) {
  row_acl result;
  char *tt;
  int u, g, o, i, d, who;

  u = g = o = i = d = 0;
  who = 1;
  result = (a) ? a : ROW_ACL_NOTHING;
  tt = VARDATA(t);
  for (i = 0; i < VARSIZE(t)-VARHDRSZ; i++) {
    if (who) {
      switch (*tt) {
        case 'u': u = 1; break;
        case 'g': g = 1; break;
        case 'o': o = 1; break;
        case 'a': u = 1; g = 1; o = 1; break;
        default: who = 0;
      }
    }
    if (!who) {
      switch (*tt) {
        case '+': d = 1; break;
        case '-': d = -1; break;
        case 'r':
          if (d > 0) {
            if (u) result |= ROW_ACL_USER_READ;
            if (g) result |= ROW_ACL_GROUP_READ;
            if (o) result |= ROW_ACL_OTHER_READ;
          }
          if (d < 0) {
            if (u) result ^= ROW_ACL_USER_READ;
            if (g) result ^= ROW_ACL_GROUP_READ;
            if (o) result ^= ROW_ACL_OTHER_READ;
          }
          break;
        case 'w':
          if (d > 0) {
            if (u) result |= ROW_ACL_USER_WRITE;
            if (g) result |= ROW_ACL_GROUP_WRITE;
            if (o) result |= ROW_ACL_OTHER_WRITE;
          }
          if (d < 0) {
            if (u) result ^= ROW_ACL_USER_WRITE;
            if (g) result ^= ROW_ACL_GROUP_WRITE;
            if (o) result ^= ROW_ACL_OTHER_WRITE;
          }
          break;
        case 'd':
          if (d > 0) {
            if (u) result |= ROW_ACL_USER_DELETE;
            if (g) result |= ROW_ACL_GROUP_DELETE;
            if (o) result |= ROW_ACL_OTHER_DELETE;
          }
          if (d < 0) {
            if (u) result ^= ROW_ACL_USER_DELETE;
            if (g) result ^= ROW_ACL_GROUP_DELETE;
            if (o) result ^= ROW_ACL_OTHER_DELETE;
          }
          break;
        case 'i':
          if (d > 0) {
            if (u) result |= ROW_ACL_USER_INSERT;
            if (g) result |= ROW_ACL_GROUP_INSERT;
            if (o) result |= ROW_ACL_OTHER_INSERT;
          }
          if (d < 0) {
            if (u) result ^= ROW_ACL_USER_INSERT;
            if (g) result ^= ROW_ACL_GROUP_INSERT;
            if (o) result ^= ROW_ACL_OTHER_INSERT;
          }
          break;
        case 'x':
          if (d > 0) {
            if (u) result |= ROW_ACL_USER_EXEC;
            if (g) result |= ROW_ACL_GROUP_EXEC;
            if (o) result |= ROW_ACL_OTHER_EXEC;
          }
          if (d < 0) {
            if (u) result ^= ROW_ACL_USER_EXEC;
            if (g) result ^= ROW_ACL_GROUP_EXEC;
            if (o) result ^= ROW_ACL_OTHER_EXEC;
          }
          break;
        case 'a':
          if (d > 0) {
            if (u) result |= ROW_ACL_USER_ALL;
            if (g) result |= ROW_ACL_GROUP_ALL;
            if (o) result |= ROW_ACL_OTHER_ALL;
          }
          if (d < 0) {
            if (u) result ^= ROW_ACL_USER_ALL;
            if (g) result ^= ROW_ACL_GROUP_ALL;
            if (o) result ^= ROW_ACL_OTHER_ALL;
          }
          break;
        default: u = g = o = d = 0; who = 1;
      }
    }
    tt++;
  }
  return result;
}

Datum
gist_row_acl_compress(PG_FUNCTION_ARGS) {
  PG_RETURN_POINTER(PG_GETARG_POINTER(0));
}

Datum
gist_row_acl_decompress(PG_FUNCTION_ARGS) {
  PG_RETURN_POINTER(PG_GETARG_POINTER(0));
}

Datum
gist_row_acl_same(PG_FUNCTION_ARGS) {
  row_acl* a = (row_acl*)PG_GETARG_POINTER(0);
  row_acl* b = (row_acl*)PG_GETARG_POINTER(1);
  bool* result =  (bool*)PG_GETARG_POINTER(2);

  if (a && b)
    *result = (*a == *b) ? TRUE : FALSE;
  else
    *result = (a == NULL && b == NULL) ? TRUE : FALSE;
  PG_RETURN_POINTER(result);
}

static int
row_acl_count_ones(row_acl a) {
  int i, v = 0;
  for (i = 0; i < ROW_ACL_COUNT; i++)
    v += (a >> i) & 1;
  return(v);
}

Datum
gist_row_acl_penalty(PG_FUNCTION_ARGS) {
  /*  ACL */
  GISTENTRY *a  = (GISTENTRY*)PG_GETARG_POINTER(0);
  /*  ACL */
  GISTENTRY *b  = (GISTENTRY*)PG_GETARG_POINTER(1);
  float *result =     (float*)PG_GETARG_POINTER(2);

  row_acl* o = (row_acl*)DatumGetPointer(a->key);

  /*   ACL */
  row_acl u = *((row_acl*)DatumGetPointer(a->key)) | *((row_acl*)DatumGetPointer(b->key));

  /*  "ߣ"   "ߣ"  ACL */
  *result = (float)(row_acl_count_ones(u) - row_acl_count_ones(*o));

  PG_RETURN_POINTER(result);
}

Datum
gist_row_acl_union(PG_FUNCTION_ARGS) {
  GistEntryVector *ev = (GistEntryVector*)PG_GETARG_POINTER(0);
  int *sizep = (int*)PG_GETARG_POINTER(1);
  int i, n = ev->n;
  row_acl* u = (row_acl*)palloc(sizeof(row_acl));

  *u = ROW_ACL_NOTHING;
  for (i = 0; i < n; i++)
    *u |= *((row_acl*)DatumGetPointer(ev->vector[i].key));

  /*      */
  *sizep = sizeof(row_acl);

  PG_RETURN_POINTER(u);
}

Datum
gist_row_acl_consistent(PG_FUNCTION_ARGS)
{
  GISTENTRY *entry = (GISTENTRY*)PG_GETARG_POINTER(0);
  row_acl   *query =   (row_acl*)PG_GETARG_POINTER(1);
  StrategyNumber strategy = PG_GETARG_UINT16(2);
  row_acl* a;
  bool result = FALSE;

  if (DatumGetPointer(entry->key) == NULL || query == NULL)
    PG_RETURN_BOOL(FALSE);

  a = (row_acl*)DatumGetPointer(entry->key);

  /*      ,    ,
   *      true    
   *  ()   
   */
  switch (strategy) {
    case RTOverlapStrategyNumber:
      result = ((*a & *query) != ROW_ACL_NOTHING) ? TRUE : FALSE;;
      break;
    case RTSameStrategyNumber:
      if (GIST_LEAF(entry))
	result = (*a == *query) ? TRUE : FALSE;
      else
	result = ((*a & *query) == *query) ? TRUE : FALSE;
      break;
    case RTContainsStrategyNumber:
      result = ((*a & *query) == *query) ? TRUE : FALSE;
      break;
    case RTContainedByStrategyNumber:
      if (GIST_LEAF(entry))
	result = ((*a & *query) == *a) ? TRUE : FALSE;
      else
	result = ((*a & *query) != ROW_ACL_NOTHING) ? TRUE : FALSE;
      break;
    default:
      result = FALSE;
  }
  PG_RETURN_BOOL(result);
}

Datum
gist_row_acl_picksplit(PG_FUNCTION_ARGS) {
  GistEntryVector *ev = (GistEntryVector*)PG_GETARG_POINTER(0);
  GIST_SPLITVEC   *v  =   (GIST_SPLITVEC*)PG_GETARG_POINTER(1);
  OffsetNumber i, maxoff = ev->n - 1;
  int n;

  n = (maxoff + 2) * sizeof(OffsetNumber);
  v->spl_left   = (OffsetNumber*)palloc(n);
  v->spl_right  = (OffsetNumber*)palloc(n);
  v->spl_ldatum = PointerGetDatum(palloc(sizeof(row_acl)));
  v->spl_rdatum = PointerGetDatum(palloc(sizeof(row_acl)));
  v->spl_nleft  = 0;
  v->spl_nright = 0;

  for (i = FirstOffsetNumber; i <= maxoff; i = OffsetNumberNext(i)) {
    row_acl* u; /*    ACL */
    row_acl* a = (row_acl*)DatumGetPointer(ev->vector[i].key);

//  if (row_acl_count_ones(*a) > 7) {
    if (i%2 == 0) {
      v->spl_left[v->spl_nleft] = i;
      v->spl_nleft++;
      u = (row_acl*)DatumGetPointer(v->spl_ldatum);
    } else {
      v->spl_right[v->spl_nright] = i;
      v->spl_nright++;
      u = (row_acl*)DatumGetPointer(v->spl_rdatum);
    }

    if (i <= OffsetNumberNext(FirstOffsetNumber)) {
      /*    ACL */
      *u = *((row_acl*)DatumGetPointer(ev->vector[i].key));
    } else {
      *u |= *a;
    }
  }
  PG_RETURN_POINTER(v);
}
